"""
Visualization methods.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging

import numpy as np
import sklearn.decomposition as skd
import sklearn.manifold as skm

import eta.core.utils as etau

import fiftyone.core.brain as fob
import fiftyone.core.utils as fou
import fiftyone.core.validation as fov
import fiftyone.zoo as foz

from fiftyone.brain.visualization import (
    VisualizationResults,
    UMAPVisualizationConfig,
    TSNEVisualizationConfig,
    PCAVisualizationConfig,
    ManualVisualizationConfig,
)
import fiftyone.brain.internal.core.utils as fbu

umap = fou.lazy_import("umap")


logger = logging.getLogger(__name__)


_DEFAULT_MODEL = "mobilenet-v2-imagenet-torch"
_DEFAULT_BATCH_SIZE = None


def compute_visualization(
    samples,
    patches_field,
    embeddings,
    points,
    brain_key,
    num_dims,
    method,
    model,
    force_square,
    alpha,
    batch_size,
    num_workers,
    skip_failures,
    **kwargs,
):
    """See ``fiftyone/brain/__init__.py``."""

    fov.validate_collection(samples)

    if method == "manual" and points is None:
        raise ValueError(
            "You must provide your own `points` when `method='manual'`"
        )

    if points is not None:
        method = "manual"
        model = None
        embeddings = None
        embeddings_field = None
        num_dims = points.shape[1]
    elif model is None and embeddings is None:
        model = foz.load_zoo_model(_DEFAULT_MODEL)
        if batch_size is None:
            batch_size = _DEFAULT_BATCH_SIZE

    if etau.is_str(embeddings):
        embeddings_field = embeddings
        embeddings = None
    else:
        embeddings_field = None

    config = _parse_config(
        embeddings_field, model, patches_field, method, num_dims, **kwargs
    )

    brain_method = config.build()
    brain_method.ensure_requirements()

    if brain_key is not None:
        brain_method.register_run(samples, brain_key)

    if points is None:
        embeddings = fbu.get_embeddings(
            samples,
            model=model,
            patches_field=patches_field,
            embeddings_field=embeddings_field,
            embeddings=embeddings,
            force_square=force_square,
            alpha=alpha,
            batch_size=batch_size,
            num_workers=num_workers,
            skip_failures=skip_failures,
        )

        logger.info("Generating visualization...")
        points = brain_method.fit(embeddings)

    results = VisualizationResults(samples, config, points)
    brain_method.save_run_results(samples, brain_key, results)

    return results


class Visualization(fob.BrainMethod):
    def ensure_requirements(self):
        pass

    def fit(self, embeddings):
        raise NotImplementedError("subclass must implement fit()")

    def get_fields(self, samples, brain_key):
        fields = []
        if self.config.patches_field is not None:
            fields.append(self.config.patches_field)

        return fields

    def cleanup(self, samples, brain_key):
        pass


class UMAPVisualization(Visualization):
    def ensure_requirements(self):
        fou.ensure_package(
            "umap-learn>=0.5",
            error_msg=(
                "You must install the `umap-learn>=0.5` package in order to "
                "use UMAP-based visualization. This is recommended, as UMAP "
                "is awesome! If you do not wish to install UMAP, try "
                "`method='tsne'` instead"
            ),
        )

    def fit(self, embeddings):
        _umap = umap.UMAP(
            n_components=self.config.num_dims,
            n_neighbors=self.config.num_neighbors,
            metric=self.config.metric,
            min_dist=self.config.min_dist,
            random_state=self.config.seed,
            verbose=self.config.verbose,
        )
        return _umap.fit_transform(embeddings)


class TSNEVisualization(Visualization):
    def fit(self, embeddings):
        if self.config.pca_dims is not None:
            _pca = skd.PCA(
                n_components=self.config.pca_dims,
                svd_solver=self.config.svd_solver,
                random_state=self.config.seed,
            )
            embeddings = _pca.fit_transform(embeddings)

        embeddings = embeddings.astype(np.float32, copy=False)

        verbose = 2 if self.config.verbose else 0

        _tsne = skm.TSNE(
            n_components=self.config.num_dims,
            perplexity=self.config.perplexity,
            learning_rate=self.config.learning_rate,
            metric=self.config.metric,
            init="pca",  # "random" or "pca"
            n_iter=self.config.max_iters,
            random_state=self.config.seed,
            verbose=verbose,
        )
        return _tsne.fit_transform(embeddings)


class PCAVisualization(Visualization):
    def fit(self, embeddings):
        _pca = skd.PCA(
            n_components=self.config.num_dims,
            svd_solver=self.config.svd_solver,
            random_state=self.config.seed,
        )
        return _pca.fit_transform(embeddings)


class ManualVisualization(Visualization):
    def fit(self, embeddings):
        raise NotImplementedError(
            "The low-dimensional representation must be manually provided "
            "when using this method"
        )


def _parse_config(
    embeddings_field, model, patches_field, method, num_dims, **kwargs
):
    if method is None or method == "umap":
        config_cls = UMAPVisualizationConfig
    elif method == "tsne":
        config_cls = TSNEVisualizationConfig
    elif method == "pca":
        config_cls = PCAVisualizationConfig
    elif method == "manual":
        config_cls = ManualVisualizationConfig
    else:
        raise ValueError("Unsupported method '%s'" % method)

    return config_cls(
        embeddings_field=embeddings_field,
        model=model,
        patches_field=patches_field,
        num_dims=num_dims,
        **kwargs,
    )



"""
Visualization tests.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import unittest

import cv2
import numpy as np

import fiftyone as fo
import fiftyone.brain as fob
import fiftyone.zoo as foz


_DATASET_NAME = "visualization-test"


def test_mnist():
    dataset = foz.load_zoo_dataset("mnist", split="test")

    # pylint: disable=no-member
    embeddings = np.array(
        [
            cv2.imread(f, cv2.IMREAD_UNCHANGED).ravel()
            for f in dataset.values("filepath")
        ]
    )

    results = fob.compute_visualization(
        dataset, embeddings=embeddings, num_dims=2, verbose=True, seed=51
    )

    plot = results.visualize(labels="ground_truth.label")
    plot.show()

    input("Press enter to continue...")


def test_images():
    dataset = _load_dataset()

    results = fob.compute_visualization(
        dataset, embeddings="embeddings", num_dims=2, verbose=True, seed=51
    )

    plot = results.visualize(labels="uniqueness")
    plot.show()

    input("Press enter to continue...")


def test_objects():
    dataset = _load_dataset()

    results = fob.compute_visualization(
        dataset,
        patches_field="ground_truth",
        embeddings="gt_embeddings",
        num_dims=2,
        verbose=True,
        seed=51,
    )

    plot = results.visualize(labels="ground_truth.detections.label")
    plot.show()

    input("Press enter to continue...")


def test_objects_subset():
    dataset = _load_dataset()

    results = fob.compute_visualization(
        dataset,
        patches_field="ground_truth",
        embeddings="gt_embeddings",
        num_dims=2,
        verbose=True,
        seed=51,
    )

    counts = dataset.count_values("ground_truth.detections.label")
    classes = sorted(counts, key=counts.get, reverse=True)[:5]

    plot = results.visualize(
        labels="ground_truth.detections.label", classes=classes
    )
    plot.show()

    input("Press enter to continue...")


def _load_dataset():
    if fo.dataset_exists(_DATASET_NAME):
        return fo.load_dataset(_DATASET_NAME)

    dataset = foz.load_zoo_dataset("quickstart", dataset_name=_DATASET_NAME)
    dataset.persistent = True

    model = foz.load_zoo_model("inception-v3-imagenet-torch")

    # Embed images
    dataset.compute_embeddings(
        model, embeddings_field="embeddings", batch_size=16
    )

    # Embed ground truth patches
    dataset.compute_patch_embeddings(
        model,
        "ground_truth",
        embeddings_field="gt_embeddings",
        batch_size=16,
        force_square=True,
    )

    return dataset


if __name__ == "__main__":
    fo.config.show_progress_bars = True
    unittest.main(verbosity=2)



"""
Utilities.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging

import numpy as np

import eta.core.utils as etau

import fiftyone.core.patches as fop
import fiftyone.zoo as foz
from fiftyone import ViewField as F


logger = logging.getLogger(__name__)


def get_ids(samples, patches_field=None):
    if patches_field is None:
        sample_ids = np.array(samples.values("id"))
        return sample_ids, None

    sample_ids = []
    label_ids = []
    for l in samples._get_selected_labels(fields=patches_field):
        sample_ids.append(l["sample_id"])
        label_ids.append(l["label_id"])

    return np.array(sample_ids), np.array(label_ids)


def filter_ids(view, samples, sample_ids, label_ids, patches_field=None):
    # No filtering required
    if view == samples or view.view() == samples.view():
        return view, sample_ids, label_ids, None

    if patches_field is None:
        _sample_ids = view.values("id")
        keep_inds = _get_keep_inds(_sample_ids, sample_ids)
        return view, np.array(_sample_ids), None, keep_inds

    # Filter labels in patches view

    if (
        isinstance(view, fop.PatchesView)
        and patches_field != view.patches_field
    ):
        raise ValueError(
            "This patches view contains labels from field '%s', not "
            "'%s'" % (view.patches_field, patches_field)
        )

    if isinstance(view, fop.EvaluationPatchesView) and patches_field not in (
        view.gt_field,
        view.pred_field,
    ):
        raise ValueError(
            "This evaluation patches view contains patches from "
            "fields '%s' and '%s', not '%s'"
            % (view.gt_field, view.pred_field, patches_field)
        )

    labels = view._get_selected_labels(fields=patches_field)
    _sample_ids = [l["sample_id"] for l in labels]
    _label_ids = [l["label_id"] for l in labels]
    keep_inds = _get_keep_inds(_label_ids, label_ids)
    return view, np.array(_sample_ids), np.array(_label_ids), keep_inds


def _get_keep_inds(ids, ref_ids):
    if len(ids) == len(ref_ids) and list(ids) == list(ref_ids):
        return None

    inds_map = {_id: idx for idx, _id in enumerate(ref_ids)}

    keep_inds = []
    bad_ids = []
    for _id in ids:
        ind = inds_map.get(_id, None)
        if ind is not None:
            keep_inds.append(ind)
        else:
            bad_ids.append(_id)

    num_bad = len(bad_ids)

    if num_bad == 1:
        raise ValueError(
            "The provided view contains ID '%s' not present in the index"
            % bad_ids[0]
        )

    if num_bad > 1:
        raise ValueError(
            "The provided view contains %d IDs (eg '%s') not present in the "
            "index" % (num_bad, bad_ids[0])
        )

    return np.array(keep_inds, dtype=np.int64)


def get_embeddings(
    samples,
    model=None,
    patches_field=None,
    embeddings_field=None,
    embeddings=None,
    force_square=False,
    alpha=None,
    handle_missing="skip",
    agg_fcn=None,
    batch_size=None,
    num_workers=None,
    skip_failures=True,
):
    if model is not None:
        if etau.is_str(model):
            model = foz.load_zoo_model(model)

        if patches_field is not None:
            logger.info("Computing patch embeddings...")
            embeddings = samples.compute_patch_embeddings(
                model,
                patches_field,
                embeddings_field=embeddings_field,
                force_square=force_square,
                alpha=alpha,
                handle_missing=handle_missing,
                batch_size=batch_size,
                num_workers=num_workers,
                skip_failures=skip_failures,
            )
        else:
            logger.info("Computing embeddings...")
            embeddings = samples.compute_embeddings(
                model,
                embeddings_field=embeddings_field,
                batch_size=batch_size,
                num_workers=num_workers,
                skip_failures=skip_failures,
            )
    elif embeddings_field is not None:
        embeddings = samples.values(embeddings_field)

    if embeddings is None:
        raise ValueError(
            "One of `model`, `embeddings_field`, or `embeddings` must be "
            "provided"
        )

    if isinstance(embeddings, dict):
        embeddings = [
            embeddings.get(_id, None) for _id in samples.values("id")
        ]

    if patches_field is not None:
        _handle_missing_patch_embeddings(embeddings, samples, patches_field)

        if agg_fcn is not None:
            embeddings = [agg_fcn(e) for e in embeddings]
            embeddings = np.stack(embeddings)
        else:
            embeddings = np.concatenate(embeddings, axis=0)
    else:
        _handle_missing_embeddings(embeddings)

        if agg_fcn is not None:
            embeddings = [agg_fcn(e) for e in embeddings]

        embeddings = np.stack(embeddings)

    return embeddings


def _handle_missing_embeddings(embeddings):
    if isinstance(embeddings, np.ndarray):
        return

    missing_inds = []
    num_dims = None
    for idx, embedding in enumerate(embeddings):
        if embedding is None:
            missing_inds.append(idx)
        elif num_dims is None:
            num_dims = embedding.size

    if not missing_inds:
        return

    missing_embedding = np.zeros(num_dims or 16)
    for idx in missing_inds:
        embeddings[idx] = missing_embedding.copy()

    logger.warning("Using zeros for %d missing embeddings", len(missing_inds))


def _handle_missing_patch_embeddings(embeddings, samples, patches_field):
    missing_inds = []
    num_dims = None
    for idx, embedding in enumerate(embeddings):
        if embedding is None:
            missing_inds.append(idx)
        elif num_dims is None:
            num_dims = embedding.shape[1]

    if not missing_inds:
        return

    missing_embedding = np.zeros(num_dims or 16)

    _, labels_path = samples._get_label_field_path(patches_field)
    patch_counts = samples.values(F(labels_path).length())

    num_missing = 0
    for idx in missing_inds:
        count = patch_counts[idx]
        embeddings[idx] = np.tile(missing_embedding, (count, 1))
        num_missing += count

    if num_missing > 0:
        logger.warning(
            "Using zeros for %d missing patch embeddings", num_missing
        )



"""
Uniqueness tests.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import os
import unittest

import eta.core.storage as etas
import eta.core.utils as etau

import fiftyone as fo
import fiftyone.brain as fob
import fiftyone.zoo as foz


def test_uniqueness():
    _run_uniqueness()


def test_uniqueness_torch_model():
    model = foz.load_zoo_model("inception-v3-imagenet-torch")
    _run_uniqueness(model=model, batch_size=16)


def test_uniqueness_tf_model():
    model = foz.load_zoo_model("resnet-v2-50-imagenet-tf1")
    _run_uniqueness(model=model, batch_size=16)


def test_roi_uniqueness():
    _run_uniqueness(roi_field="ground_truth")


def test_roi_uniqueness_torch_model():
    model = foz.load_zoo_model("inception-v3-imagenet-torch")
    _run_uniqueness(roi_field="ground_truth", model=model, batch_size=16)


def test_roi_uniqueness_tf_model():
    model = foz.load_zoo_model("resnet-v2-50-imagenet-tf1")
    _run_uniqueness(roi_field="ground_truth", model=model, batch_size=16)


def _run_uniqueness(roi_field=None, model=None, batch_size=None):
    dataset = foz.load_zoo_dataset(
        "quickstart", dataset_name=fo.get_default_dataset_name()
    )
    dataset.delete_sample_field("uniqueness")

    view = dataset.take(50)
    num_samples = len(view)

    # Custom Torch model
    fob.compute_uniqueness(
        view, roi_field=roi_field, model=model, batch_size=batch_size
    )

    num_uniqueness = dataset.count("uniqueness")
    assert num_uniqueness == num_samples

    bounds = dataset.bounds("uniqueness")

    assert bounds[0] >= 0
    assert bounds[1] <= 1


if __name__ == "__main__":
    fo.config.show_progress_bars = True
    unittest.main(verbosity=2)
